from financial_canvas.figures.constants import DAY_IN_MILLIS
from financial_canvas.figures.utils import read_file

from bokeh.models import (DateRangeSlider, CustomJS)


class FilteringDateRangeSlider(object):
    '''
    Does not have any sources, not a figure even (TODO: make separate module?), 
        updates sources and x range of the target figure.

    Args:
        target_figure (figures.CustomFigure): (optional) if passed will be resized with slider.

    Attributes:
        s (bokeh.models.DateRangeSlider): 
    '''

    def __init__(self, target_figure=None):

        preselected_index = target_figure.sources['main'][0].data['index']
        all_data_index = target_figure.sources['main'][1].data['index']
        date_range_slider = DateRangeSlider(
            title="Select period to render",
            value=(preselected_index[0], preselected_index[-1]),
            start=all_data_index[0],
            end=all_data_index[-1],
            step=DAY_IN_MILLIS,
        )

        resizable_sources = list(target_figure.sources.values())
        date_range_slider.js_on_change("value", CustomJS(
            args=dict(
                sources_and_origins=resizable_sources,
                x_range=target_figure.p.x_range,
            ),
            code=read_file('update_render_range.js')))

        self.s = date_range_slider



from functools import partial

from bokeh.plotting import figure as bokeh_figure
from bokeh.models import CustomJS
from financial_canvas.figures.utils import create_sources
from financial_canvas.figures.utils import read_file


class CustomFigure(object):
    '''
    A base class for creating bokeh figures that will filter. Bokeh 
    already has live data filtering capabilities through CDSViews but it has its drawbacks:
        * lines are not supported by filtered view bokeh, related bokeh issues:
            * https://github.com/bokeh/bokeh/issues/9388
            * https://github.com/bokeh/bokeh/issues/7070
        * y axis autoscale does not work with zoom
            * TODO: report to bokeh

    Sources and figure are extendable so that it's possible add new elements to the figure.

    Args:
        df (pandas.DataFrame): should have 'date_time' as index 
            (in pandas.DatetimeIndex format), main sources will be created from this df
            all columns will be added to sources
        file_name (str): Error code.
        file_name (int, optional): Error code.
        selected_from (pandas.Timestamp) date from which to select the short initially drawn source

    Attributes:
        sources (dict with tuples with bokeh.models.ColumnDataSource (source, origin)): 
            source - initial source that the chart will be made from 
                (short, before any js callback updates) 
            origin - large source, will be used to recreate source after inside JS callbacks
        p (bokeh.models.Figure): the plot with glyphs
        selected_from (pandas.Timestamp) date from which to select
    '''

    def __init__(self, df, *, selected_from=None):
        if selected_from is None:
            selected_from = df.index[0]
        self.sources = create_sources(df, selected_from=selected_from)
        self.selected_from = selected_from
        self.p = None

    def add_hover(self, columns):
        hover_tool = self.p.hover
        # hover_tool.tooltips = []
        # hover_tool.formatters = {}
        for column_name, pretty_name in columns.items():
            hover_tool.tooltips.extend([
                (pretty_name, '@' + column_name + '{%f}'),
            ])
            hover_tool.formatters.update({
                '@' + column_name: 'printf',
            })

    def get_figure_defaults(self):
        return partial(
            bokeh_figure,
            # even though "webgl" is advertised as the fastest engine
            # not using it here because it introduces a bug (TODO: report bug to bokeh repo)
            # when panning/zooming the window some glyphs like circles or diamonds can stay
            # where from previous view
            output_backend="canvas",
            # TODO: pass arguments to constructor
            plot_height=450,
            margin=(10, 10, 10, 10),
            height_policy='fixed',
            sizing_mode='stretch_width',
        )

    def y_axis_autorange(self):
        yaxis = self.p.left[0]
        yaxis.formatter.use_scientific = False

        if (self.y_range_resize_columns):
            y_axis_auto_range_callback = CustomJS(
                args=dict(
                    unique_name=id(self),
                    y_range=self.p.y_range,
                    # TODO: deal with multiple sources
                    source=self.sources['main'][0],
                    columns=self.y_range_resize_columns,
                ),
                code=read_file('y_axis_auto_range.js')
            )
            self.p.x_range.js_on_change('start', y_axis_auto_range_callback)
            self.p.x_range.js_on_change('end', y_axis_auto_range_callback)

    def add_sources(self, df, name):
        additional_sources = create_sources(
            df, selected_from=self.selected_from, name=name)
        self.sources.update(additional_sources)
        return self.sources



from financial_canvas.figures.CustomFigure import CustomFigure

from bokeh.models import RangeTool


class PreviewSlider(CustomFigure):
    '''
    Creates preview slider at the bottom of the main figure. The first df should always contain
    close column.

    Args:
        target_figure (figures.CustomFigure): (optional) if passed will be resized with slider.
    '''

    def __init__(self, df, *, selected_from=None, target_figure=None):
        super().__init__(df, selected_from=selected_from)

        bokeh_figure = self.get_figure_defaults()

        p = bokeh_figure(
            title="Drag the middle and edges of the selection box to change the range above",
            plot_height=125,
            x_axis_type="datetime",
            y_axis_type=None,
            tools="", toolbar_location=None,
            background_fill_color="#efefef"
        )

        if target_figure:
            range_tool = RangeTool(x_range=target_figure.p.x_range)
            range_tool.overlay.fill_color = "navy"
            range_tool.overlay.fill_alpha = 0.2
            p.add_tools(range_tool)
            p.toolbar.active_multi = range_tool
            self.sources = target_figure.sources

        p.scatter(
            'index', 'close',
            size=1,
            source=self.sources['main'][0],
        )

        # layout
        p.ygrid.grid_line_color = None

        self.p = p




from bokeh.models import DatetimeTickFormatter
from bokeh.models.tools import HoverTool
from financial_canvas.figures.constants import COLOR_PALLETE
from financial_canvas.figures.CustomFigure import CustomFigure
from bokeh.models import Span


class Line(CustomFigure):
    '''
    Simple line chart with y axis autorange
    '''

    def __init__(self, df, *, column_name, figure_args, selected_from=None, color=None, x_range=None):

        # TODO: update dynamically
        self.y_range_resize_columns = [column_name]

        super().__init__(df, selected_from=selected_from)

        tooltips = [
            ('Date', '@index{%d/%m/%Y  %H:%M}'),
            (column_name, '@' + column_name + '{%f}'),
        ]

        formatters = {
            '@index': 'datetime',
            '@' + column_name: 'printf',
        }

        # using only sources (not origins) to build initial graphs
        # if switching to bokeh CDSView needs to be changed
        full_source = self.sources['main'][0]
        bokeh_figure = self.get_figure_defaults()

        x_range = x_range if x_range else (
            full_source.data['index'][0], full_source.data['index'][-1])
        p = bokeh_figure(
            x_axis_type="datetime",
            # x_axis_location="above",
            toolbar_location='right',
            tools="pan,wheel_zoom,box_zoom,save",
            active_drag='pan',
            active_scroll='wheel_zoom',
            # need to set custom x_range here, because otherwise
            # the preview slider will crash with the following error:
            # ValueError: expected an instance of type Range1d, got DataRange1d(id='1006', ...) of type DataRange1d
            # also used to set up initial zoom if passed
            x_range=x_range,
            **figure_args,
        )

        p.toolbar.logo = None

        ticks_formatter = DatetimeTickFormatter(
            years=["%Y"],
            days=["%d/%m/%Y"],
            months=["%m/%Y"],
            hours=["%H:%M"],
            minutes=["%H:%M"]
        )

        p.add_tools(HoverTool(
            tooltips=tooltips,
            formatters=formatters,
            mode='vline',
            names=['line'],
        ))

        color = color if color else COLOR_PALLETE[2]

        # zero horizontal line
        zero_hline = Span(location=0,
                          dimension='width', line_color='gray',
                          line_dash='dashed', line_width=1)
        p.add_layout(zero_hline)

        # line chart
        p.line('index', column_name,
               legend_label=column_name,
               name='line',
               line_color=color,
               source=full_source)

        # legend
        p.legend.location = "top_left"
        p.legend.click_policy = "hide"

        # layout
        # TODO: add to defaults
        p.xaxis.formatter = ticks_formatter
        p.axis.minor_tick_line_color = None
        p.axis.major_tick_line_color = None
        p.outline_line_color = None
        p.axis.axis_line_width = 0
        p.axis.major_tick_out = 8
        p.axis.major_tick_in = 0

        self.p = p
        self.y_axis_autorange()



from bokeh.models import ColumnDataSource

import importlib.resources as pkg_resources
from .. import js_callbacks


def create_sources(df, selected_from=None, name='main'):
    sources = {}

    def get_bokeh_source(df):
        dict_source = {'index': df.index}
        for column in df.columns:
            dict_source[column] = df[column]
        return ColumnDataSource(dict_source)

    source_df = df[df.index >= selected_from] if selected_from else df
    source = get_bokeh_source(source_df)
    origin = get_bokeh_source(df)

    sources.update({
        name: (source, origin),
    })

    return sources


def read_file(file_name):
    return pkg_resources.read_text(js_callbacks, file_name)




from datetime import timedelta

from bokeh.models import DatetimeTickFormatter
from bokeh.models.tools import HoverTool
from financial_canvas.figures.constants import COLOR_PALLETE
from financial_canvas.figures.CustomFigure import CustomFigure


class Candlestick(CustomFigure):
    '''
    Interactive candlestick chart on canvas with y-range autoscale.

    Args:
        df: pandas DataFrame with datetime index and OHLC columns, 
        y_range_resize_columns (list of str): columns in df to resize y_range by
    '''

    def __init__(self, df, *, selected_from=None, inc_color=None, dec_color=None, x_range=None, figure_args=None):

        # TODO: update dynamically
        self.y_range_resize_columns = ['open', 'high', 'low', 'close']

        super().__init__(df, selected_from=selected_from)

        inc = df.close > df.open
        dec = ~inc
        inc_df = df[inc]
        dec_df = df[dec]

        self.add_sources(inc_df, 'inc')
        self.add_sources(dec_df, 'dec')

        tooltips = [
            ('Date', '@index{%d/%m/%Y %H:%M}'),
            ('Open', '@open{%f}'),
            ('High', '@high{%f}'),
            ('Low', '@low{%f}'),
            ('Close', '@close{%f}'),
        ]

        formatters = {
            '@index': 'datetime',
            '@open': 'printf',
            '@high': 'printf',
            '@low': 'printf',
            '@close': 'printf',
        }

        # using only sources (not origins) to build initial graphs
        # if switching to bokeh CDSView needs to be changed
        inc_source = self.sources['inc'][0]
        dec_source = self.sources['dec'][0]
        full_source = self.sources['main'][0]
        bokeh_figure = self.get_figure_defaults()

        x_range = x_range if x_range else (
            full_source.data['index'][0], full_source.data['index'][-1])
        if figure_args is None:
            figure_args = {}
        p = bokeh_figure(
            x_axis_type="datetime",
            x_axis_location="above",
            toolbar_location='right',
            tools="pan,wheel_zoom,box_zoom,save",
            active_drag='pan',
            active_scroll='wheel_zoom',
            # need to set custom x_range here, because otherwise
            # the preview slider will crash with the following error:
            # ValueError: expected an instance of type Range1d, got DataRange1d(id='1006', ...) of type DataRange1d
            # also used to set up initial zoom if passed
            x_range=x_range,
            **figure_args
        )

        p.toolbar.logo = None
        p.x_range.min_interval = timedelta(days=1)

        inc_color = inc_color if inc_color else COLOR_PALLETE[5]
        dec_color = dec_color if dec_color else COLOR_PALLETE[0]

        ticks_formatter = DatetimeTickFormatter(
            years=["%Y"],
            days=["%d/%m/%Y"],
            months=["%m/%Y"],
            hours=["%H:%M"],
            minutes=["%H:%M"]
        )

        # prep data source
        # Open - High - Low - Close
        width_ms = dict(day=86400,
                        hour=3600,
                        minute=60,
                        second=1)[inc_source.data['index'].resolution] * 1000 * .85

        self.candle_width = width_ms

        p.add_tools(HoverTool(
            tooltips=tooltips,
            formatters=formatters,
            mode='vline',
            names=['inc_bar', 'dec_bar'],
        ))

        # candlestick chart
        p.segment('index', 'high', 'index', 'low',
                  legend_label='OHLC',
                  color=dec_color,
                  source=dec_source)
        p.segment('index', 'high', 'index', 'low',
                  legend_label='OHLC',
                  color=inc_color,
                  source=inc_source)

        p.vbar('index', width_ms, 'open', 'close',
               legend_label='OHLC',
               name='dec_bar',
               fill_color=dec_color, line_color=dec_color,
               source=dec_source)
        p.vbar('index', width_ms, 'open', 'close',
               legend_label='OHLC',
               name='inc_bar',
               fill_color=inc_color, line_color=inc_color,
               source=inc_source)

        # legend
        p.legend.location = "top_left"
        p.legend.click_policy = "hide"

        # layout
        # TODO: add to defaults
        p.xaxis.formatter = ticks_formatter
        p.axis.minor_tick_line_color = None
        p.axis.major_tick_line_color = None
        p.outline_line_color = None
        p.axis.axis_line_width = 0
        p.axis.major_tick_out = 8
        p.axis.major_tick_in = 0

        self.p = p
        self.y_axis_autorange()



'''
Glyphs for indicators drawn on top of Candlestick chart
'''


from financial_canvas.figures.constants import COLOR_PALLETE, DAY_IN_MILLIS
from bokeh.models import LabelSet, CustomJS


def up_traingle(candlestick, y_column_name='open', color='green'):
    candlestick.y_range_resize_columns.extend([y_column_name])
    source = candlestick.sources['main'][0]
    # TODO: fix width
    candlestick.p.triangle(
        x='index',
        y=y_column_name,
        # width=width,
        # size=width,
        size=10,
        color=color,
        source=source,
        legend_label=y_column_name,
    )


def down_triangle(candlestick, y_column_name='open', color='red'):
    candlestick.y_range_resize_columns.extend([y_column_name])
    source = candlestick.sources['main'][0]
    # TODO: fix width
    # width = candlestick.candle_width * 0.25
    # candlestick.p.diamond(
    #     x='index',
    #     y=y_column_name,
    #     width=width,
    #     source=source,
    #     legend_label=y_column_name,
    # )

    candlestick.p.inverted_triangle(
        x='index',
        y=y_column_name,
        # width=width,
        # size=width,
        size=10,
        color=color,
        source=source,
        legend_label=y_column_name,
    )


def annotations(candlestick, column_name, y_column_name="open"):
    source = candlestick.sources['main'][0]
    renderer = candlestick.p.rect(
        color='black',
        legend_label='BUY/SELL',
    )

    labels = LabelSet(
        x='index', y=y_column_name,
        text=column_name,
        x_offset=0, y_offset=10,
        source=source, render_mode='canvas',
        background_fill_color='white',
        text_baseline='middle',
        text_align="center",
    )
    candlestick.p.add_layout(labels)
    renderer.js_on_change('visible', CustomJS(args=dict(ls=labels),
                                              code="ls.visible = cb_obj.visible;"))


def line(custom_figure, column_name, color=None):
    color = color if color else COLOR_PALLETE[2]
    custom_figure.y_range_resize_columns.extend([column_name])

    custom_figure.p.line(
        'index', column_name,
        legend_label=column_name,
        color=color,
        line_width=1.5,
        source=custom_figure.sources['main'][0],
    )

    hover_tool = custom_figure.p.hover
    hover_tool.tooltips.extend([
        (column_name, '@' + column_name + '{%f}'),
    ])
    hover_tool.formatters.update({
        '@' + column_name: 'printf',
    })


def line_circles(candlestick, column_name, color=None):
    color = color if color else COLOR_PALLETE[2]
    candlestick.y_range_resize_columns.extend([column_name])

    width_ms = dict(day=86400,
                        hour=3600,
                        minute=60,
                        second=1)[candlestick.sources['main'][0].data['index'].resolution] * 500 * .85

    candlestick.p.scatter(
        'index', column_name,
        legend_label=column_name,
        radius=width_ms,
        color=color,
        source=candlestick.sources['main'][0]
    )

    candlestick.p.line(
        'index', column_name,
        legend_label=column_name,
        color=color,
        line_width=1.5,
        source=candlestick.sources['main'][0],
    )

    hover_tool = candlestick.p.hover
    hover_tool.tooltips.extend([
        (column_name, '@' + column_name + '{%f}'),
    ])
    hover_tool.formatters.update({
        '@' + column_name: 'printf',
    })


def line_ma_st(candlestick):
    ma_color = COLOR_PALLETE[4]
    st_color = COLOR_PALLETE[1]
    candlestick.y_range_resize_columns.extend(['MA', 'ST'])

    candlestick.p.scatter(
        'index', 'MA',
        legend_label='MA',
        radius=750000,
        color=ma_color,
        source=candlestick.sources['main'][0]
    )

    candlestick.p.scatter(
        'index', 'ST',
        legend_label='ST',
        radius=750000,
        color=st_color,
        source=candlestick.sources['main'][0]
    )

    candlestick.p.line(
        'index', 'MA',
        legend_label='MA',
        color=ma_color,
        source=candlestick.sources['main'][0],
    )
    candlestick.p.line(
        'index', 'ST',
        legend_label='ST',
        color=st_color,
        source=candlestick.sources['main'][0],
    )

    hover_tool = candlestick.p.hover
    hover_tool.tooltips.extend([
        ('MA', '@MA{%f}'),
        ('ST', '@ST{%f}'),
    ])
    hover_tool.formatters.update({
        '@MA': 'printf',
        '@ST': 'printf',
    })


def strategy_indicators(candlestick, line_column, circles_column):
    # Indicator 1 (the orange line) is just a continuous line
    indicator_1_color = COLOR_PALLETE[2]

    candlestick.y_range_resize_columns.extend([line_column])

    candlestick.p.line(
        'index', line_column,
        legend_label=line_column,
        color=indicator_1_color,
        line_width=1.5,
        source=candlestick.sources['main'][0],
    )

    candlestick.p.circle(
        'index', f'{circles_column}_position',
        legend_label=circles_column,
        # radius=1000 * 60 * 15,
        size=5,
        color=f'{circles_column}_color',
        source=candlestick.sources['main'][0],
    )

    hover_tool = candlestick.p.hover
    hover_tool.tooltips.extend([
        (line_column, '@' + line_column + '{%f}'),
        (circles_column, '@' + circles_column + '{%f}'),
    ])
    hover_tool.formatters.update({
        f'@{line_column}': 'printf',
        f'@{circles_column}': 'printf',
    })


def strategy_indicators_linear(candlestick, line_column, colored_column, df):
    # Indicator 1 (the orange line) is just a continuous line
    indicator_1_color = COLOR_PALLETE[2]

    candlestick.y_range_resize_columns.extend([line_column, colored_column])

    candlestick.p.line(
        'index', line_column,
        legend_label=line_column,
        color=indicator_1_color,
        line_width=1.5,
        source=candlestick.sources['main'][0],
    )

    # if making differently colored lines:
    # we cannot use a single or 2 line glyphs because
    # the line is not connected
    # it's either multi_line https://docs.bokeh.org/en/latest/docs/reference/models/glyphs/multi_line.html
    # but then we will need to rethink date range slicing
    # ~~or it's multiple different sources for each consecutive continuous line~~ (nope, this is not working out at all,
    # very slow build and there is a json parsing error in the browser)
    # or a segment https://docs.bokeh.org/en/latest/docs/reference/models/glyphs/segment.html
    # but with segment it's not clear how to draw a point of one
    candlestick.p.line(
        'index', colored_column,
        legend_label=colored_column,
        color='#c0c0c0',
        line_width=1.5,
        source=candlestick.sources['main'][0],
    )

    # also adding circles for lines of length == 1
    candlestick.p.circle(
        'index', colored_column,
        legend_label=colored_column,
        # line_width=1.5,
        size=5,
        color=f'{colored_column}_color',
        source=candlestick.sources['main'][0],
    )

    hover_tool = candlestick.p.hover
    hover_tool.tooltips.extend([
        (line_column, '@' + line_column + '{%f}'),
        (colored_column, '@' + colored_column + '{%f}'),
    ])
    hover_tool.formatters.update({
        f'@{line_column}': 'printf',
        f'@{colored_column}': 'printf',
    })




from bokeh.models import DatetimeTickFormatter
from bokeh.models.tools import HoverTool
from financial_canvas.figures.constants import COLOR_PALLETE
from financial_canvas.figures.CustomFigure import CustomFigure
from bokeh.models import Span, Scatter
import seaborn as sns


class MultiCircle(CustomFigure):
    '''
    A figure with multiple circles. Each cirlce color corresponds to each column name
    docs for the form of the points: https://docs.bokeh.org/en/latest/docs/reference/models/glyphs/scatter.html#bokeh.models.glyphs.Scatter
    '''

    def __init__(self, df, columns, *, selected_from=None, figure_args=None, scatter_args=None, colors=None, x_range=None):
        # TODO: update dynamically
        self.y_range_resize_columns = columns

        if figure_args is None:
            figure_args = {}

        if scatter_args is None:
            scatter_args = {}

        if selected_from is None:
            selected_from = df.index[0]

        if colors is None:
            colors = sns.color_palette("colorblind", len(columns)).as_hex()

        super().__init__(df, selected_from=selected_from)

        tooltips = [
            ('Date', '@index{%d/%m/%Y  %H:%M}'),
        ]

        formatters = {
            '@index': 'datetime',
        }

        for column_name in columns:
            tooltips.append((column_name, '@' + column_name + '{%f}'))
            formatters.update({'@' + column_name: 'printf'})

        # using only sources (not origins) to build initial graphs
        # if switching to bokeh CDSView needs to be changed
        full_source = self.sources['main'][0]
        bokeh_figure = self.get_figure_defaults()

        x_range = x_range if x_range else (
            full_source.data['index'][0], full_source.data['index'][-1])
        p = bokeh_figure(
            x_axis_type="datetime",
            # x_axis_location="above",
            toolbar_location='right',
            tools="pan,wheel_zoom,box_zoom,save",
            active_drag='pan',
            active_scroll='wheel_zoom',
            # need to set custom x_range here, because otherwise
            # the preview slider will crash with the following error:
            # ValueError: expected an instance of type Range1d, got DataRange1d(id='1006', ...) of type DataRange1d
            # also used to set up initial zoom if passed
            x_range=x_range,
            **figure_args,
        )

        p.toolbar.logo = None

        ticks_formatter = DatetimeTickFormatter(
            years=["%Y"],
            days=["%d/%m/%Y"],
            months=["%m/%Y"],
            hours=["%H:%M"],
            minutes=["%H:%M"]
        )

        p.add_tools(HoverTool(
            tooltips=tooltips,
            formatters=formatters,
            mode='vline',
            names=[columns[0]],
        ))

        # zero horizontal line
        zero_hline = Span(location=0,
                          dimension='width', line_color='gray',
                          line_dash='dashed', line_width=1)
        p.add_layout(zero_hline)

        radius_ms = dict(day=86400,
                        hour=3600,
                        minute=60,
                        second=1)[df.index.resolution] * 1000 * .85
        for column, color in zip(columns, colors):
            # TODO: make resizable circles on zoom
            # specifying radius breaks onhover behavor
            updated_scatter_args = dict(
                x='index', y=column,
                legend_label=column,
                name=column,
                color=color,
                size=6,
                source=full_source,
            )
            updated_scatter_args.update(scatter_args)
            p.circle(
                **updated_scatter_args,
            )

        # legend
        p.legend.location = "top_left"
        p.legend.click_policy = "hide"

        # layout
        # TODO: add to defaults
        p.xaxis.formatter = ticks_formatter
        p.axis.minor_tick_line_color = None
        p.axis.major_tick_line_color = None
        p.outline_line_color = None
        p.axis.axis_line_width = 0
        p.axis.major_tick_out = 8
        p.axis.major_tick_in = 0

        self.p = p
        self.y_axis_autorange()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("requirements.txt", "r") as fh:
    requirements = [line.strip() for line in fh]

setuptools.setup(
    name="financial_canvas",
    version="1.3.2",
    author="Evgeniia Vakarina",
    author_email="27793901+EvgeniiaVak@users.noreply.github.com",
    description="Draw interactive candlesticks on browser canvas from Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=requirements,
)